#lang scheme

(require redex/reduction-semantics)

(provide/contract
 [can-reach (->* (reduction-relation? any/c) (#:max-terms (or/c number?)) #:rest (listof any/c) any/c)]
 [can-reach/pred (->* ((-> any/c (listof any/c)) any/c) (#:max-terms (or/c number?)) #:rest (listof any/c) any/c)]
 [can-reach/pred/1+ (->* ((-> any/c (listof any/c)) any/c) (#:max-terms (or/c number?)) #:rest (listof any/c) any/c)]
 [path-from-to/reversed (-> reduction-relation? any/c any/c (listof any/c))])
(provide (struct-out exn:search-failure))

(define-struct (exn:search-failure exn:fail) ())

;; can the transitive closure of the relation get from t1 to the t2s? (uses breadth-first search)
;; reduces to can-reach/pred
(define (can-reach relation #:max-terms [max-terms #f] t1 . t2s)
  (let ([target-preds (for/list ([t2 (in-list t2s)])
                        (lambda (t) (equal? t t2)))])
    (apply can-reach/pred
           (λ (t) (apply-reduction-relation relation t))
           t1 target-preds #:max-terms max-terms)))

;; can the transitive closure of the relation get from t1 to terms satisfying the preds?
(define (can-reach/pred #:max-terms [max-terms #f] next-steps t1 . t2-preds)
  (let ([first-table (add-to-table! (make-hash) (list t1))])
    (breadth-first-search first-table hash-has-key? add-to-table! next-steps (list t1) t2-preds #:max-terms max-terms)))

;; like can-reach/pred, but requires at least one step.
(define (can-reach/pred/1+  #:max-terms [max-terms #f] next-steps t1 . preds)
  (let* ([newly-reachable (next-steps t1)]
         [unsatisfied-preds (filter (lambda (pred) (not (ormap pred newly-reachable))) preds)]
         [first-table (add-to-table! (make-hash) newly-reachable)])
    (breadth-first-search first-table hash-has-key? add-to-table! next-steps newly-reachable preds #:max-terms max-terms)))


;; add-to-table: given a hash and some new keys, add them to the table and return it.
(define (add-to-table! hash new-keys)
  (for ([k (in-list new-keys)])
    (hash-set! hash k #t))
  hash)


;; path-from-to : return a *reversed* reduction sequence going from a to b, using the relation.
;; another breadth-first seach, associating the live set with the set of states used to reach it
(define (path-from-to/reversed next-steps from to)
  ;; the path-table maps elements to the shortest path that reaches that element
  (let* ([path-table (make-hash)]
         [_1 (hash-set! path-table from (list from))]
         [visited-table (add-to-table! (make-hash) from)]
         [adjacent-to (lambda (t)
                        (let* ([path-to-t (hash-ref path-table t)]
                               [reachable (next-steps t)])
                          (for ([r (in-list reachable)])
                            (unless (hash-has-key? path-table r)
                              (hash-set! path-table r (cons r path-to-t))))
                          reachable))]
         [pred (lambda (t) (equal? t to))])
    (breadth-first-search visited-table hash-has-key? add-to-table! adjacent-to (list from) (list pred))
    ;; the path-table must now contain the shortest path to the desired node.
    (hash-ref path-table to)))



;; breadth-first-search, abstract in a number of ways. Accepts a representation of the nodes 
;; visited, a procedure that checks in that representation, a 

;; breadth-first-search : table (table element -> boolean) (table (listof element) -> table) (listof element) (listof (element -> boolean)) -> boolean
;; signals an exception if max-terms is exceeded.
(define (breadth-first-search visited v-check update-visited adjacent-to frontier target-preds #:max-terms [max-terms #f])
  (let loop ([visited visited] 
             [frontier frontier]
             [target-preds (remove* frontier target-preds (lambda (a b) (b a)))]
             [max-terms max-terms])
    (cond [(and max-terms (<= max-terms 0))
           (raise (make-exn:search-failure "max search terms exceeded" (current-continuation-marks)))]
          [else
           (let ([visited? (lambda (t) (v-check visited t))])
             (or (empty? target-preds)
                 (and (not (empty? frontier))
                      (let* ([newly-reachable (filter (lambda (t) (not (visited? t))) (adjacent-to (first frontier)))]
                             [new-visited (update-visited visited newly-reachable)])
                        (loop new-visited
                              (append (rest frontier) newly-reachable) 
                              (remove* newly-reachable target-preds (lambda (a b) (b a)))
                              (and max-terms (- max-terms 1)))))))])))
